@@ -45,7 +45,7 @@
 #include "harness.h"
 
 #if __TBB_GCC_WARNING_SUPPRESSION_PRESENT
-#pragma GCC diagnostic ignored "-Wuninitialized"
+//#pragma GCC diagnostic ignored "-Wuninitialized"
 #endif
 
 static tbb::atomic<int> construction_counter;
